/* $Id: uds.c,v 1.4 2002/06/14 13:54:04 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include "util.h"
#include "dstring.h"
#include "lists.h"
#include "interp.h"
#define UDS_IMP
#include "uds.h"

storage_class uds_storage_class = {uds_create, uds_destroy, 
									 uds_check, uds_get,
									 uds_put, uds_add,
									 uds_remove,
									 uds_first, uds_next,
									 uds_enditer,
									 uds_numofitems,
									 uds_seek
									};

typedef struct uds_data
{
	dstring *id;
	interpreter *handler;
} uds_data;

storage *uds_create(interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	value *val;
	
	ns->data = ALLOC(uds_data);
	((uds_data *)ns->data)->id = ds_fix(ds_create(name));
	((uds_data *)ns->data)->handler = find_subint(i, arg);
	ns->class = &uds_storage_class;
	val = ALLOC(value);
	val->val = ds_fix(((uds_data *)ns->data)->id);
	discard_value(eval_procedure(((uds_data *)ns->data)->handler, "create", val));
	return ns;
}

static value *get_storage_name(uds_data *uds)
{
	value *nv = ALLOC(value);
	nv->val = ds_fix(uds->id);
	return nv;
}

void uds_destroy(interpreter *i, storage *s)
{
	uds_data *ud = s->data;
	discard_value(eval_procedure(ud->handler, "destroy", get_storage_name(ud)));
}

int uds_check(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value);
	dstring *ret;
	arg->val = ds_fix(key);
	ret = eval_procedure(ud->handler, "check", list_add(arg, get_storage_name(ud)));
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret);
}

dstring *uds_get(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value);
	dstring *ret;
	arg->val = ds_fix(key);
	ret = eval_procedure(ud->handler, "get", list_add(arg, get_storage_name(ud)));
	if(ret)
		ds_unfix(ret);
	return ret ? ret : ds_create(NULL);
}

void uds_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value), *arg1 = ALLOC(value);
	arg->val = ds_fix(key);
	arg1->val = ds_fix(val);
	discard_value(eval_procedure(ud->handler, "put", list_add(list_add(arg1, arg), get_storage_name(ud))));
}

void uds_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value), *arg1 = ALLOC(value);
	arg->val = ds_fix(key);
	arg1->val = ds_fix(val);
	discard_value(eval_procedure(ud->handler, "insert", list_add(list_add(arg1, arg), get_storage_name(ud))));
}

void uds_remove(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value);
	arg->val = ds_fix(key);
	discard_value(eval_procedure(ud->handler, "remove", list_add(arg, get_storage_name(ud))));
}

static int unique_uds_enum_id = 1;
dstring *uds_first(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	dstring *ret;
	value *arg = ALLOC(value);
	arg->val = ds_fix(ds_fromint(unique_uds_enum_id, 10, 0));
	*(int *)iter = unique_uds_enum_id++;
	ret = eval_procedure(ud->handler, "first", list_add(arg, get_storage_name(ud)));
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

dstring *uds_next(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	dstring *ret;
	value *arg = ALLOC(value);
	arg->val = ds_fix(ds_fromint(*(int *)iter, 10, 0));
	ret = eval_procedure(ud->handler, "next", list_add(arg, get_storage_name(ud)));
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

void uds_enditer(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	value *arg = ALLOC(value);
	arg->val = ds_fix(ds_fromint(*(int *)iter, 10, 0));
	discard_value(eval_procedure(ud->handler, "enditer", list_add(arg, get_storage_name(ud))));
}

int uds_numofitems(interpreter *i, storage *s)
{
	uds_data *ud = s->data;
	dstring *ret;
	ret = eval_procedure(ud->handler, "count", get_storage_name(ud));
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? convert_to_int(DS_BODY(ret)) : 0;
}

dstring *uds_seek(interpreter *i, storage *s, void **iter, dstring *key)
{
	uds_data *ud = s->data;
	dstring *ret;
	value *arg = ALLOC(value), *arg1 = ALLOC(value);
	arg->val = ds_fix(ds_fromint(unique_uds_enum_id, 10, 0));
	arg1->val = ds_fix(key);
	*(int *)iter = unique_uds_enum_id++;
	ret = eval_procedure(ud->handler, "seek", list_add(list_add(arg1, arg), get_storage_name(ud)));
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

void uds_init(interpreter *i, storage *s, value *args)
{
	uds_data *ud = s->data;
	discard_value(eval_procedure(ud->handler, "init", list_add(copy_values(args), get_storage_name(ud))));
}
